"""This module contains version information of the library.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.__version__

    As example.

    Also, `VersionInfo` isn't imports by default. You can't use it with code like this:

    .. code-block:: python

        pyrolog.VersionInfo

    See below for details how to use it, if you really want this.
"""

from collections import namedtuple

__all__ = ['__version__', '__version_tuple__', '__version_info__']

VersionInfo = namedtuple('VersionInfo', ['major', 'minor', 'patch', 'release_level', 'commit'])
"""Named tuple type.

.. warning::
    This isn't imports by default! You can import this by using this code:
    
    .. code-block:: python
    
        from pyrolog.version import VersionInfo
"""

__version__ = "2.2.1"
"""String with the current version of the library"""

__version_tuple__ = (2, 2, 1)
"""Tuple with the current version of the library"""

__version_info__ = VersionInfo(*__version_tuple__, 'beta', '50f233c')
"""Named tuple with the current version + technical details of the library release"""




"""
    Pyrolog. Pretty logging library.
    Copyright (C) 2023  ftdot (https://github.com/ftdot)
"""

from . import defaults
from . import _types as types
from . import utils
from . import empty_colors

from ._types import LogOnlyLevels

from .group import *
from .logger import *
from .logging_context import *
from .handlers import *
from .formatters import *
from .version import *
from .colors import *


def get_plain_logger(log_level: types.LogLevel = 'info'):
    """Gets logger (unnamed) with StdoutHandler and plain formatter. Also, uses given log level.

    :param log_level: Log level
    :type log_level: types.LogLevel
    """

    return Logger(
        handlers=[StdoutHandler(log_level=log_level), ]
    )


def get_colored_logger(log_level: types.LogLevel = 'info'):
    """Gets logger (unnamed) with StdoutHandler and colored formatter. Also, uses given log level.

    :param log_level: Log level
    :type log_level: types.LogLevel
    """

    return Logger(
        handlers=[StdoutHandler(log_level=log_level, formatter=ColoredFormatter()), ]
    )



"""Contains base of formatters and all defined in the library formatters.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.PlainFormatter

    As example.
"""

import traceback
import datetime

from abc import abstractmethod
from collections import namedtuple
from functools import lru_cache

from . import empty_colors
from .logging_context import LoggingContext
from ._types import VarDict, ColorDict
from .defaults import (DEFAULT_LOGGING_CONTEXT,
                       MINIMAL_FORMAT_STRING,
                       MINIMAL_TIME_FORMAT_STRING,
                       COLORED_MINIMAL_FORMAT_STRING,
                       COLORED_MINIMAL_TIME_FORMAT_STRING,
                       DEFAULT_COLOR_DICT)
from .colors import TextColor, BGColor, TextStyle

from typing import Any

__all__ = ['fmt', 'Uncolored', 'Formatter', 'PlainFormatter', 'ColoredFormatter']

fmt = namedtuple('FormatTuple', ('format_string', 'string'))
"""Special type to specify formatting in :class:`ColoredFormatter` use case."""

####


class Uncolored:
    """Used in context where required to log an object to the :class:`ColoredFormatter` without colors.

    :ivar value: Original value.
    :type value: Any
    """

    def __init__(self, value: Any):
        """
        :param value: Any value you want to pass to the :class:`ColoredFormatter` without colors.
        """
        self.value = value

    def __format__(self, format_spec):
        return self.value.__format__(format_spec)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return repr(self.value)

####


class Formatter:
    """A base of all the formatters.

    :ivar format_string: Format string.
    :type format_string: str
    :ivar time_format_string: Format string of the time.
    :type time_format_string: str
    :ivar static_variables: Static variables. (can be used in messages)
    :type static_variables: VarDict | None
    :ivar logging_context: Logging context.
    :type logging_context: LoggingContext
    :ivar time_formatting: (**System variable.** Do not change it manually) Determines whether to spend time formatting time.
    :type time_formatting: bool
    """

    def __init__(self,
                 format_string: str = MINIMAL_FORMAT_STRING,
                 time_format_string: str = MINIMAL_TIME_FORMAT_STRING,
                 static_variables: VarDict | None = None,
                 logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT
                 ):
        """
        :param format_string: Format string.
        :type format_string: str
        :param time_format_string: Format string of the time.
        :type time_format_string: str
        :param static_variables: Static variables. (can be used in messages)
        :type static_variables: VarDict | None
        :param logging_context: Logging context.
        :type logging_context: LoggingContext
        """

        self._format_string      = format_string
        self.time_format_string  = time_format_string
        self.static_variables    = {} if static_variables is None else static_variables
        self.logging_context     = logging_context

        self.time_formatting = '{time}' in self.format_string

        defined_formatters.append(self)

    @property
    def format_string(self):
        return self._format_string

    @format_string.setter
    def format_string(self, value: str):
        self._format_string   = value
        self.time_formatting  = '{time}' in value

    def add_static_variable(self, name: str, value: Any):
        """Adds static variable.

        :param name: Name of the static variable.
        :type name: str
        :param value: Value of the static variable.
        :type value: Any
        """

        self.static_variables[name] = value

    def del_static_variable(self, name: str) -> bool:
        """Deletes static variable.

        :param name: Name of the static variable to be deleted.
        :type name: str
        """

        if name in self.static_variables:
            del self.static_variables[name]
            return True
        return False

    @abstractmethod
    def format(self,
               message: str,
               time: datetime.datetime | None,
               level: str,
               logger_color: str,
               logger_name: str,
               group_name: str,
               group_color: str,
               fmt_args: list[Any],
               fmt_kwargs: dict[str, Any]
               ):
        raise NotImplementedError('Method "format()" isn\'t implemented')

    @abstractmethod
    def format_exception(self, exc: Exception):
        raise NotImplementedError('Method "format_exception()" isn\'t implemented')

    @abstractmethod
    def format_time(self, time: datetime.datetime | None):
        raise NotImplementedError('Method "format_time()" isn\'t implemented')


class PlainFormatter(Formatter):
    """Plain formatter with offsets support.

    :ivar offsets: Determines whether to use offsets or not.
    :type offsets: bool
    """

    def __init__(self, *args: Any, offsets: bool = True, **kwargs: dict[str, Any]):
        """
        :param offsets: If True, format string can use offsets to prettify output.
        :type offsets: bool
        """
        super().__init__(*args, **kwargs)

        self.offsets                                 = offsets
        self.static_variables['level_offset']        = self.logging_context.get_level_offset() if offsets else 0
        self.static_variables['logger_name_offset']  = self.logging_context.get_logger_name_offset() if offsets else 0
        self.static_variables['group_name_offset']   = self.logging_context.get_group_name_offset() if offsets else 0
        self.static_variables['fore']                = empty_colors.EmptyTextColor
        self.static_variables['bg']                  = empty_colors.EmptyBGColor
        self.static_variables['style']               = empty_colors.EmptyTextStyle
        self.static_variables['reset']               = ''

    def format_message(self,
                       message: str,
                       level: str,
                       logger_color: str,
                       logger_name: str,
                       group_name: str,
                       group_color: str,
                       fmt_args: list[Any],
                       fmt_kwargs: dict[str, Any]):
        return message.format(
            level=level,
            logger_color=logger_color,
            logger_name=logger_name,
            group_name=group_name,
            group_color=group_color,
            *fmt_args,
            **fmt_kwargs,
            **self.static_variables,
        )

    def format(self,
               message: str,
               time: datetime.datetime | None,
               level: str,
               logger_color: str,
               logger_name: str,
               group_name: str,
               group_color: str,
               fmt_args: list[Any],
               fmt_kwargs: dict[str, Any]
               ):
        return self.format_string.format(
            message=self.format_message(
                message,
                level,
                logger_color,
                logger_name,
                group_name,
                group_color,
                fmt_args,
                fmt_kwargs
            ),
            time=self.format_time(time) if self.time_formatting else '*',
            level=level,
            logger_color=logger_color,
            logger_name=logger_name,
            group_name=group_name,
            group_color=group_color,
            *fmt_args,
            **fmt_kwargs,
            **self.static_variables,
        )

    def format_exception(self, exc: Exception):
        return ''.join(traceback.format_exception(exc))[:-1]

    def format_time(self, time: datetime.datetime | None):
        if time is None:
            return ''

        return self.time_format_string.format(
            year=time.year,
            month=time.month,
            day=time.day,
            hour=time.hour,
            minute=time.minute,
            second=time.second,
            microsecond=str(time.microsecond)[:6].ljust(6),
        )


class ColoredFormatter(PlainFormatter):
    """Colored formatter.

    :ivar color_dict: Dict with the colors used to color formatting arguments by its type.
    :type color_dict: ColorDict
    :ivar use_repr: If it set to True, repr() will be used instead of str() while format arguments.
    :type use_repr: bool
    """

    def __init__(self,
                 format_string: str = COLORED_MINIMAL_FORMAT_STRING,
                 time_format_string: str = COLORED_MINIMAL_TIME_FORMAT_STRING,
                 *args: Any,
                 color_dict: ColorDict = DEFAULT_COLOR_DICT,
                 use_repr: bool = False,
                 **kwargs: dict[str, Any]):
        """
        :param format_string: Format string.
        :type format_string: str
        :param time_format_string: Format string of the time.
        :type time_format_string: str
        :param color_dict: Dict with the colors used to color formatting arguments by its type.
        :type color_dict: ColorDict
        :param use_repr: If it set to True, repr() will be used instead of str() while format arguments.
        :type use_repr: bool
        """
        super().__init__(format_string, time_format_string, *args, **kwargs)

        self.color_dict  = color_dict
        self.use_repr    = use_repr

        self.static_variables['fore']          = TextColor
        self.static_variables['bg']            = BGColor
        self.static_variables['style']         = TextStyle
        self.static_variables['reset']         = TextStyle.reset

        self._func = repr if use_repr else str

    def get_level_color(self, level: str) -> str:
        if level in self.color_dict['levels']:
            return self.color_dict['levels'][level]
        else:
            return ''

    def get_value_color(self, type_: type) -> str:
        if type_ in self.color_dict['types']:
            return self.color_dict['types'][type_]
        elif isinstance(type_, Exception):
            return self.color_dict['types']['exception']
        else:
            return self.color_dict['types']['all']

    def format_value(self, value: Any) -> str:
        if isinstance(value, fmt):
            value_color = self.get_value_color(type(value[1]))
            return value_color + value[0].format(value[1])

        elif isinstance(value, list):
            list_color = self.get_value_color(list)
            output = f'{list_color}['
            temp = [self.format_value(v) for v in value]
            output += f'{list_color}, '.join(temp) + f'{list_color}]{TextStyle.reset}'

            return output

        elif isinstance(value, tuple):
            tuple_color = self.get_value_color(tuple)
            output = f'{tuple_color}('
            temp = [self.format_value(v) for v in value]
            output += f'{tuple_color}, '.join(temp) + f'{tuple_color}){TextStyle.reset}'

            return output

        elif isinstance(value, dict):
            dict_color = self.get_value_color(dict)
            output = f'{dict_color}{{'
            temp = [f'{self.format_value(k)}{dict_color}: {self.format_value(v)}' for k, v in value.items()]
            output += f'{dict_color}, '.join(temp) + f'{dict_color}}}{TextStyle.reset}'

            return output

        elif isinstance(value, Uncolored):
            return self._func(value) + TextStyle.reset

        return self.get_value_color(type(value)) + self._func(value) + TextStyle.reset

    def format_message(self,
                       message: str,
                       level: str,
                       level_color: str,
                       logger_color: str,
                       logger_name: str,
                       group_name: str,
                       group_color: str,
                       fmt_args: list[Any],
                       fmt_kwargs: dict[str, Any]) -> str:
        return message.format(
            level=level,
            level_color=level_color,
            logger_color=logger_color,
            logger_name=logger_name,
            group_name=group_name,
            group_color=group_color,
            *fmt_args,
            **fmt_kwargs,
            **self.static_variables,
        )

    def format(self,
               message: str,
               time: datetime.datetime | None,
               level: str,
               logger_color: str,
               logger_name: str,
               group_name: str,
               group_color: str,
               fmt_args: list[Any],
               fmt_kwargs: dict[str, Any]
               ) -> str:
        colored_args    = [self.format_value(a) for a in fmt_args]
        colored_kwargs  = {k: self.format_value(v) for k, v in fmt_kwargs.items()}
        level_color     = self.get_level_color(level)

        return self.format_string.format(
            message=self.format_message(
                message,
                level,
                level_color,
                logger_color,
                logger_name,
                group_name,
                group_color,
                colored_args,
                colored_kwargs
            ),
            time=self.format_time(time) if self.time_formatting else '*',
            level=level,
            level_color=level_color,
            logger_color=logger_color,
            logger_name=logger_name,
            group_name=group_name,
            group_color=group_color,
            *colored_args,
            **colored_kwargs,
            **self.static_variables,
        ) + TextStyle.reset

    def format_exception(self, exc: Exception):
        return self.color_dict['types']['exception'] + ''.join(traceback.format_exception(exc))[:-1]

    def format_time(self, time: datetime.datetime | None):
        if time is None:
            return ''

        return self.time_format_string.format(
            year=time.year,
            month=time.month,
            day=time.day,
            hour=time.hour,
            minute=time.minute,
            second=time.second,
            microsecond=time.microsecond,  # str(time.microsecond)[:4]
            **self.static_variables,
        )

defined_formatters: list[Formatter] = []
"""List with the defined formatters."""



"""Dedicated module for :class:`Group` class.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.Group

    As example.
"""

from .handlers import Handler
from .logging_context import LoggingContext
from .logger import Logger
from .defaults import DEFAULT_LOGGING_CONTEXT
from .utils import update_group_name_offset

__all__ = ['Group']


class Group:
    """A group class, that very helpful if you use many loggers in your project.

    :ivar name: Name of the group.
    :type name: str
    :ivar handlers: Handlers to be used by loggers.
    :type handlers: list[Handler]
    :ivar logging_context: The current logging context. (by default is `defaults.DEFAULT_LOGGING_CONTEXT` of
        :mod:`pyrolog.defaults` module)
    :type logging_context: LoggingContext
    :ivar group_color: Color of the group. (visible only by using ColoredFormatter)
    :type group_color: str
    :ivar enabled: If it is set to False, all pinned loggers will not log any messages and all subgroups pinned to
        will be disabled. Do not change it manually, use :meth:`Group.disable()` and :meth:`Group.enable()`.
    :type enabled: bool
    :ivar parent_group: Parent group of this group.
    :type parent_group: Group | None
    :ivar subgroups: Subgroups of this group.
    :type subgroups: list[Group]
    :ivar loggers: Loggers pinned to this group.
    :type loggers: list[Logger]
    """

    def __init__(self,
                 name: str = '',
                 handlers: Handler | list[Handler] | None = None,
                 logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT,
                 group_color: str = '',
                 enabled: bool = True,
                 parent_group: 'Group | str | None' = None
                 ):
        """
        :param name: Name of the group.
        :type name: str
        :param handlers: Handlers to be used by loggers.
        :type handlers: Handler | list[Handler] | None
        :param logging_context: The current logging context. (by default is defaults.DEFAULT_LOGGING_CONTEXT)
        :type logging_context: LoggingContext
        :param group_color: Color of the group. (visible only by using ColoredFormatter)
        :type group_color: str
        :param enabled: If it is set to False, all pinned loggers will not log any messages and all subgroups pinned to
            will be disabled.
        :type enabled: bool
        :param parent_group: Parent of this group. If it is not None, then copies all parameters from that group.
        :type parent_group: Group | str | None
        """

        if parent_group is None:
            self.handlers = [] if handlers is None else (
                [handlers, ] if isinstance(handlers, Handler) else handlers
            )
            self.logging_context  = logging_context
            self.group_color      = group_color
            self.enabled          = enabled
            self.name_path        = name

        else:
            if isinstance(parent_group, str):
                if parent_group not in logging_context.groups_by_name:
                    raise NameError(f'Group "{parent_group}" isn\'t defined in given logging context.')

                parent_group = logging_context.groups_by_name[parent_group]

            self.handlers         = parent_group.handlers
            self.logging_context  = parent_group.logging_context
            self.enabled          = parent_group.enabled
            self.name_path        = parent_group.name_path + '.' + name
            self.group_color      = parent_group.group_color if group_color == '' else group_color

            parent_group.subgroups.append(self)
        self.name                     = name
        self.subgroups: list[Group]   = []
        self.loggers: list['Logger']  = []
        self.parent_group             = parent_group

        self.logging_context.groups.append(self)
        self.logging_context.groups_by_name[name] = self
        update_group_name_offset(self.logging_context)

    def enable(self):
        """Enables this group and all pinned loggers and subgroups."""

        self.enabled = True

        for l in self.loggers:
            l.enable()

        for sg in self.subgroups:
            sg.enable()

    def disable(self):
        """Enables this group and all pinned loggers and subgroups."""

        self.enabled = False

        for l in self.loggers:
            l.disable()

        for sg in self.subgroups:
            sg.disable()

    def subgroup(self, *args, **kwargs):
        """Alternative constructor of the :class:`Group`. But with "parent_group" set to exist instance.
        This method is shorthand for `Group(..., parent_group=self)`. All arguments passed to this function
        is passed to the constructor of the group.

        Example:

        .. code-block:: python

            example_group      = pyrolog.Group('ExampleGroup', handlers=[..., ])
            example_sub_group  = example_group.subgroup()
            example_logger     = example_group.group(name='ExampleLogger', group_color=pyrolog.TextColor.red)
        """
        return Group(*args, **kwargs, parent_group=self)

    def logger(self, *args, **kwargs):
        """Alternative constructor of the :class:`Logger`. But with "group" set to exist instance.
        This method is shorthand for `Logger(..., group=self)`. All arguments passed to this function
        is passed to the constructor of the logger.

        Example:

        .. code-block:: python

            example_group   = pyrolog.Group('ExampleGroup', handlers=[..., ])
            example_logger  = example_group.logger(name='ExampleLogger', logger_color=pyrolog.TextColor.cyan)
        """
        return Logger(*args, **kwargs, group=self)



"""Contains empty version of the color bindings of the colorama library. See :mod:`pyrolog.colors`.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.empty_colors.EmptyTextColor

    As example.
"""


class EmptyTextColor:
    """Empty variant of :class:`pyrolog.TextColor`."""

    reset         = ''
    black         = ''
    red           = ''
    green         = ''
    blue          = ''
    cyan          = ''
    yellow        = ''
    magenta       = ''
    white         = ''
    lightblack    = ''
    lightred      = ''
    lightgreen    = ''
    lightblue     = ''
    lightcyan     = ''
    lightyellow   = ''
    lightmagenta  = ''
    lightwhite    = ''


class EmptyBGColor:
    """Empty variant of :class:`pyrolog.BGColor`."""

    reset         = ''
    black         = ''
    red           = ''
    green         = ''
    blue          = ''
    cyan          = ''
    yellow        = ''
    magenta       = ''
    white         = ''
    lightblack    = ''
    lightred      = ''
    lightgreen    = ''
    lightblue     = ''
    lightcyan     = ''
    lightyellow   = ''
    lightmagenta  = ''
    lightwhite    = ''


class EmptyTextStyle:
    """Empty variant of :class:`pyrolog.TextStyle`."""

    reset   = ''
    dim     = ''
    bold    = ''
    normal  = ''



"""Dedicated module for the :class:`Logger` class.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.Logger

    As example.
"""

from datetime import datetime

from .handlers import Handler
from .utils import make_logger_binding, update_logger_name_offset
from .logging_context import LoggingContext
from .defaults import DEFAULT_LOGGING_CONTEXT
from ._types import LogLevel

from typing import TYPE_CHECKING, Any, Self

if TYPE_CHECKING:
    from .group import Group

__all__ = ['Logger']


class Logger:
    """A logger object.

    :ivar name: Name of the logger.
    :type name: str
    :ivar handlers: Handlers to be used by logger.
    :type handlers: list[Handler]
    :ivar logging_context: The current logging context. (by default is defaults.DEFAULT_LOGGING_CONTEXT)
    :type logging_context: LoggingContext
    """

    def __init__(self,
                 name: str = '',
                 handlers: Handler | list[Handler] | None = None,
                 logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT,
                 logger_color: str = '',
                 group: 'Group | str | None' = None,
                 enabled: bool = True,
                 ):
        """Creates a new Logger object.

        :param name: Name of the logger. (by default it is empty)
        :type name: str | None
        :param handlers: Handlers to be used by logger.
        :type handlers: Handler | list[Handler] | None
        :param logging_context: The current logging context. (by default is defaults.DEFAULT_LOGGING_CONTEXT)
        :type logging_context: LoggingContext
        :param logger_color: Color of the logger. (visible only by using ColoredFormatter)
        :type logger_color: str
        :param group: Group of the logger. If it is not None, then copies all parameters from that group.
        :type group: Group | str | None
        :param enabled: If it is set to False, logger will not log any messages.
        :type enabled: bool
        """

        if group is None:
            self.handlers         = [handlers, ] if isinstance(handlers, Handler) else handlers
            self.logging_context  = logging_context
            self.enabled          = enabled

            self.group_name_path  = '*'
            self.group_color      = ''

        else:
            self.logging_context = logging_context if isinstance(group, str) else group.logging_context
            self.to_group(group)

        self.name          = name
        self.logger_color  = logger_color
        self.group         = group

        self.logging_context.loggers.append(self)
        update_logger_name_offset(self.logging_context)

    def to_group(self, group: 'Group | str'):
        """Moves logger to the given group.

        :param group: Group where be placed logger.
        :type group: Group | str
        """

        if isinstance(group, str):
            if group not in self.logging_context.groups_by_name:
                raise NameError(f'Group "{group}" isn\'t defined in given logging context.')

            group = self.logging_context.groups_by_name[group]

        self.handlers         = group.handlers
        self.logging_context  = group.logging_context
        self.enabled          = group.enabled

        self.group_name_path  = group.name_path
        self.group_color      = group.group_color

        group.loggers.append(self)

    def set_level(self, level: LogLevel):
        """Sets given log level to all handlers.

        :param level: Log level.
        :type level: LogLevel
        """

        for h in self.handlers:
            h.set_level(level)

    def add_handler(self, handler: Handler):
        """Adds a new handler to the logger.

        :param handler: New handler to be added.
        :type handler: Handler
        """

        self.handlers.append(handler)

    def remove_handler(self, handler: Handler):
        """Removes handler from the logger. **Ignores if handler isn't used in logger.**

        :param handler: Handler to be removed.
        :type handler: Handler
        """

        if handler in self.handlers:
            self.handlers.remove(handler)

    def enable(self):
        """Enables a logger."""
        self.enabled = True

    def disable(self):
        """Disables a logger."""
        self.enabled = False

    def record(self,
               message: str,
               level: str | int,
               *args: Any,
               exc: Exception | None = None,
               **kwargs: dict[str, Any]):
        """Records a message at the specified logging level.

        :param message: Message to be logged.
        :type message: str
        :param level: Logged level.
        :type level:
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """

        if not self.enabled:
            return

        for h in self.handlers:
            h.write(
                message,
                level,
                self.logger_color,
                self.name,
                self.group_name_path,
                self.group_color,
                exc=exc,
                time=datetime.now(),
                fmt_args=args,
                fmt_kwargs=kwargs,
            )

    @staticmethod
    def bind_log_methods():
        """Binds all log methods for the Logger object instance."""

        for level in DEFAULT_LOGGING_CONTEXT.log_levels:
            setattr(Logger, level, make_logger_binding(level))

    ####

    # These methods are implemented by "_bind_log_methods()" method (when the Logger object initialized)

    def debug(self, message: str, *args: Any, exc: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs debug message. Shorthand for :method:`Logger.record()` with `level='debug'`.

        :param message: Message to be logged.
        :type message: str
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """
        ...

    def exception(self, message: str, *args: Any, exc: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs exception message. Shorthand for :method:`Logger.record()` with `level='exception'`.

        :param message: Message to be logged.
        :type message: str
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """
        ...

    def info(self, message: str, *args: Any, exception: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs info message. Shorthand for :method:`Logger.record()` with `level='info'`.

        :param message: Message to be logged
        :type message: str
        :param args: Positioned arguments for formatting
        :type args: Any
        :param exception: Exception to pin with log message
        :type exception: Exception
        :param kwargs: Named arguments for formatting
        :type kwargs: dict[str, Any]
        """
        ...

    def warn(self, message: str, *args: Any, exc: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs warn message. Shorthand for :method:`Logger.record()` with `level='info'`.

        :param message: Message to be logged.
        :type message: str
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """
        ...

    def error(self, message: str, *args: Any, exc: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs error message. Shorthand for :method:`Logger.record()` with `level='info'`.

        :param message: Message to be logged.
        :type message: str
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """
        ...

    def critical(self, message: str, *args: Any, exc: Exception | None = None, **kwargs: dict[str, Any]):
        """Logs critical message. Shorthand for :method:`Logger.record()` with `level='info'`.

        :param message: Message to be logged.
        :type message: str
        :param args: Positioned arguments for formatting.
        :type args: Any
        :param exc: Exception to pin with log message.
        :type exc: Exception
        :param kwargs: Named arguments for formatting.
        :type kwargs: dict[str, Any]
        """
        ...

Logger.bind_log_methods()



"""Contains the color bindings to the colorama library.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.TextColor

    As example.
"""

import sys

from colorama import Fore, Back, Style, init, just_fix_windows_console

# fix windows console if current platform is windows, and fix isn't disabled
if sys.platform == 'win32' and '--pyrolog-disable-windows-fix' not in sys.argv:
    init()
    just_fix_windows_console()

__all__ = ['TextColor', 'BGColor', 'TextStyle']


class TextColor:
    """Bindings to colorama Fore colors."""

    reset         = Fore.RESET
    black         = Fore.BLACK
    red           = Fore.RED
    green         = Fore.GREEN
    blue          = Fore.BLUE
    cyan          = Fore.CYAN
    yellow        = Fore.YELLOW
    magenta       = Fore.MAGENTA
    white         = Fore.WHITE
    lightblack    = Fore.LIGHTBLACK_EX
    lightred      = Fore.LIGHTRED_EX
    lightgreen    = Fore.LIGHTGREEN_EX
    lightblue     = Fore.LIGHTBLUE_EX
    lightcyan     = Fore.LIGHTCYAN_EX
    lightyellow   = Fore.LIGHTYELLOW_EX
    lightmagenta  = Fore.LIGHTMAGENTA_EX
    lightwhite    = Fore.LIGHTWHITE_EX


class BGColor:
    """Bindings to colorama Back colors."""

    reset         = Back.RESET
    black         = Back.BLACK
    red           = Back.RED
    green         = Back.GREEN
    blue          = Back.BLUE
    cyan          = Back.CYAN
    yellow        = Back.YELLOW
    magenta       = Back.MAGENTA
    white         = Back.WHITE
    lightblack    = Back.LIGHTBLACK_EX
    lightred      = Back.LIGHTRED_EX
    lightgreen    = Back.LIGHTGREEN_EX
    lightblue     = Back.LIGHTBLUE_EX
    lightcyan     = Back.LIGHTCYAN_EX
    lightyellow   = Back.LIGHTYELLOW_EX
    lightmagenta  = Back.LIGHTMAGENTA_EX
    lightwhite    = Back.LIGHTWHITE_EX


class TextStyle:
    """Bindings to colorama Style."""

    reset   = Style.RESET_ALL
    dim     = Style.DIM
    bold    = Style.BRIGHT
    normal  = Style.NORMAL



"""This module defines types of the library.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.types.ColorDict

    As example.

    But, you can use :class:`LogOnlyLevels` as following code:

    .. code-block:: python

        pyrolog.LogOnlyLevels
"""

from functools import lru_cache

from typing import Any, TypeAlias, Literal

ColorDict: TypeAlias     = dict[Literal['types'] | Literal['levels'], dict[str | type, Any]]
VarDict: TypeAlias       = dict[str, Any]
LogLevelDict: TypeAlias  = dict[str, int]


class LogOnlyLevels:
    """A class that allows you to fine-tune the desired logging levels.

    :ivar levels: Levels to be logged.
    :type levels: list[str]
    """

    def __init__(self, levels: str | list[str]):
        """Determines which level(s) will be logged and which not.

        Example usage:

        .. code-block:: python

            import pyrolog
            stdout_handler = pyrolog.StdoutHandler(
                log_level=pyrolog.LogOnlyLevels(
                    ['debug', 'info', 'warn']
                ),
                formatter=pyrolog.PlainFormatter()
            )
            stderr_handler = pyrolog.StderrHandler(
                log_level=pyrolog.LogOnlyLevels(
                    ['exception', 'error', 'critical']
                ),
                formatter=pyrolog.PlainFormatter()
            )

        When using this handler, only the specified levels will be logged.

        :param levels: Level(s) to be logged only.
        :type levels: str | list[str]
        """

        self.levels = [levels, ] if isinstance(levels, str) else levels

    @lru_cache(10)
    def log_level(self, level: str) -> bool:
        """Checks if level is allowed to log.

        :param level: Level to check.
        :type level: str

        :returns: `True` if level is allowed to log, otherwise `False` be returned.
        :rtype: bool
        """

        return level in self.levels

LogLevel: TypeAlias = str | int | LogOnlyLevels



"""Module with the default definitions of library.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.defaults.DEFAULT_LOG_LEVELS

    As example.
"""

from .logging_context import LoggingContext
from .colors import *

DEFAULT_LOG_LEVELS = {
    'debug': 5,
    'exception': 10,
    'info': 15,
    'warn': 20,
    'error': 25,
    'critical': 30,
    'notset': 9999,
}
"""The default log levels dictionary."""

MINIMAL_FORMAT_STRING = '{level:<{level_offset}} {message}'
"""The minimal formatter specification, which is the default and recommended for small projects."""

TIMED_MINIMAL_FORMAT_STRING = '{time} {level:<{level_offset}} {message}'
"""The minimal formatter specification, but with specified time."""

MINIMAL_TIME_FORMAT_STRING = '{hour:02d}:{minute:02d}:{second:02d}.{microsecond}'
"""The default minimal time format specification."""

MAXIMUM_TIME_FORMAT_STRING = '{year:04d}.{month:02d}.{day:02d} {hour:02d}:{minute:02d}:{second:02d}.{microsecond}'
"""Time Format specification that is recommended for professional, large projects."""

MAXIMUM_TIME_FORMAT_STRING_FILENAME_SAFE = '{year:04d}_{month:02d}_{day:02d}-{hour:02d}_{minute:02d}_{second:02d}_{microsecond}'
"""Time Format specification that is recommended for filename of the log files."""

MAXIMUM_FORMAT_STRING = '{time} | {level:<{level_offset}} | {group_name:<{group_name_offset}} | {logger_name:<{logger_name_offset}} -> {message}'
"""Format specification that is recommended for professional, large projects."""

DEFAULT_LOGGING_CONTEXT = LoggingContext(DEFAULT_LOG_LEVELS)
"""The logging context that is used by all elements of the logging library by default."""

DEFAULT_COLOR_DICT = {
    'types': {
        int: TextColor.lightmagenta,
        float: TextColor.lightmagenta + TextStyle.bold,
        bool: TextColor.yellow,
        str: TextColor.lightgreen,
        bytes: TextColor.lightred + TextStyle.bold,
        list: TextColor.lightyellow,
        tuple: TextColor.lightyellow + TextStyle.bold,
        dict: TextColor.cyan,
        'exception': TextColor.lightred,
        'all': TextColor.lightcyan + TextStyle.bold,
    },
    'levels': {
        'debug': TextColor.lightwhite,
        'exception': TextColor.lightyellow,
        'info': TextColor.lightcyan,
        'warn': TextColor.yellow,
        'error': TextColor.lightred,
        'critical': TextColor.red + TextStyle.bold,
    }
}
"""The default color dict that is used by ColoredFormatter"""

COLORED_MINIMAL_FORMAT_STRING = '{level_color}{level:<{level_offset}}{reset} {message}'
"""The colored minimal formatter specification, which is the default for the colored formatter and recommended for small projects."""

COLORED_TIMED_MINIMAL_FORMAT_STRING = '{time} {level_color}{level:<{level_offset}}{reset} {message}'
"""The minimal colored formatter specification, but with specified time."""

COLORED_MINIMAL_TIME_FORMAT_STRING = '{fore.cyan}{hour:02d}{fore.reset}:{fore.cyan}{minute:02d}{fore.reset}:{fore.cyan}{second:02d}{fore.reset}.{fore.lightmagenta}{microsecond}{fore.reset}'
"""The default colored minimal time format specification."""

COLORED_MAXIMUM_TIME_FORMAT_STRING = '{fore.cyan}{year:04d}{fore.reset}.{fore.cyan}{month:02d}{fore.reset}.{fore.cyan}{day:02d}{fore.reset} {fore.cyan}{hour:02d}{fore.reset}:{fore.cyan}{minute:02d}{fore.reset}:{fore.cyan}{second:02d}{fore.reset}.{fore.lightmagenta}{microsecond}{fore.reset}'
"""Colored time format specification that is recommended for professional, large projects."""

COLORED_MAXIMUM_FORMAT_STRING = '{time} | {level_color}{level:<{level_offset}}{reset} | {group_color}{group_name:<{group_name_offset}}{reset} | {logger_color}{logger_name:<{logger_name_offset}}{reset} -> {message}'
"""Colored format specification that is recommended for professional, large projects."""



"""Module that defines a base of the handlers and all the library handlers.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.StdoutHandler

    As example.
"""

import datetime
import sys

from os import PathLike

from .formatters import Formatter, PlainFormatter
from .logging_context import LoggingContext
from .defaults import DEFAULT_LOGGING_CONTEXT
from ._types import LogLevel

from abc import abstractmethod
from typing import TextIO, Any

__all__ = ['Handler', 'IOHandler', 'StdoutHandler', 'StderrHandler', 'FileHandler']


class Handler:
    """A base of all the handlers.

    :ivar log_level: Log level.
    :type log_level: LogLevel
    :ivar formatter: Formatter that will use this handler.
    :type formatter: Formatter
    :ivar logging_context: Logging context.
    :type logging_context: LoggingContext
    :ivar log_exceptions: Determines whether log exceptions or not.
    :type log_exceptions: bool
    :ivar enabled: Determines whether log any message or not. Isn't recommended to change it manually, but you may. Is
        recommended to use :meth:`enable()` and :meth:`disable()` methods.
    :type enabled: bool
    """

    def __init__(self,
                 log_level: LogLevel = 'info',
                 formatter: Formatter = PlainFormatter(),
                 logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT,
                 log_exceptions: bool = True,
                 enabled: bool = True,
                 ):
        """
        :param log_level: Log level.
        :type log_level: LogLevel
        :param formatter: Formatter that will use this handler.
        :type formatter: Formatter
        :param logging_context: Logging context.
        :type logging_context: LoggingContext
        :param log_exceptions: Determines whether log exceptions or not.
        :type log_exceptions: bool
        :param enabled: Determines whether log any message or not.
        :type enabled: bool
        """
        self.log_level        = log_level
        self.formatter        = formatter
        self.logging_context  = logging_context
        self.log_exceptions   = log_exceptions
        self.enabled          = enabled

    def enable(self):
        """Enables handler."""

        self.enabled = True

    def disable(self):
        """Disables handler."""

        self.enabled = False

    @abstractmethod
    def write(self,
              message: str,
              level: str | int,
              logger_color: str,
              logger_name: str,
              group_name: str,
              group_color: str,
              exc: Exception | None = None,
              time: datetime.datetime | None = None,
              fmt_args: list[Any] | None = None,
              fmt_kwargs: dict[str, Any] | None = None):
        raise NotImplementedError('Method "write()" isn\'t implemented!')


class IOHandler(Handler):
    """A base of IO handlers.

    :ivar io: IO to be used to write messages.
    :type io: TextIO
    """

    def __init__(self, io: TextIO, *args: Any, **kwargs: dict[str, Any]):
        """
        :param io: IO to be used to write messages.
        :type io: TextIO
        """
        super().__init__(*args, **kwargs)

        self.io = io

    def write(self,
              message: str,
              level: str | int,
              logger_color: str,
              logger_name: str,
              group_name: str,
              group_color: str,
              exc: Exception | None = None,
              time: datetime.datetime | None = None,
              fmt_args: list[Any] | None = None,
              fmt_kwargs: dict[str, Any] | None = None):
        if not self.enabled:
            return

        if self.logging_context.log_level(self.log_level, level):

            # log formatted message
            self.io.write(self.formatter.format(
                message,
                time,
                level,
                logger_color,
                logger_name,
                group_name,
                group_color,
                fmt_args, fmt_kwargs)+'\n')

            # format and write exception to io if exceptions logging is enabled and exception was given
            if self.log_exceptions and exc is not None:
                self.io.write(self.formatter.format_exception(exc)+'\n')

            self.io.flush()

    def set_level(self, level: LogLevel):
        """Sets log level of handler to given.

        :param level: Log level.
        :type level: LogLevel
        """
        self.log_level = level


class StdoutHandler(IOHandler):
    """Handles stdout output (console).
    This class is shorthand for IOHandler(sys.stdout, ...)."""

    def __init__(self, *args: Any, **kwargs: dict[str, Any]):
        super().__init__(sys.stdout, *args, **kwargs)


class StderrHandler(IOHandler):
    """Handles stderr output (console).
    This class is shorthand for IOHandler(sys.stdout, ...)."""

    def __init__(self, *args: Any, **kwargs: dict[str, Any]):
        super().__init__(sys.stderr, *args, **kwargs)


class FileHandler(IOHandler):
    """Handles file output.

    :ivar file_io: Opened file object.
    :type file_io: TextIO
    :ivar path: Path to the file.
    :type path: str | bytes | PathLike[str] | PathLike[bytes] | int
    :ivar encoding: Encoding, by default is the UTF-8.
    :type encoding: str"""

    def __init__(self,
                 path: str | bytes | PathLike[str] | PathLike[bytes] | int,
                 encoding: str = 'utf8',
                 *args: Any,
                 **kwargs: dict[str, Any]
                 ):
        """
        :param path: Path to the file.
        :type path: str | bytes | PathLike[str] | PathLike[bytes] | int
        :param encoding: Encoding, by default is the UTF-8.
        :type encoding: str
        """
        self.file_io   = open(path, 'w', encoding=encoding)
        self.path      = path
        self.encoding  = encoding

        super().__init__(self.file_io, *args, **kwargs)

    def __del__(self):
        self.file_io.close()



"""Dedicated module for the :class:`LoggingContext` class.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.LoggingContext

    As example.
"""

from functools import lru_cache

from ._types import LogLevelDict, LogOnlyLevels, LogLevel

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .logger import Logger
    from .group import Group

__all__ = ['LoggingContext']


class LoggingContext:
    """Defines some variables that is "unique" for every logging context.

    :ivar log_levels: Dict with the registered log levels.
    :type log_levels: LogLevelDict
    :ivar loggers: List with the loggers pinned to logging context instance.
    :type loggers: list[Logger]
    :ivar groups: List with the groups pinned to logging context instance.
    :type groups: list[Group]
    :ivar groups_by_name: Dictionary with groups with names as keys.
    :type groups_by_name: dict[str, 'Group']
    """

    def __init__(self, log_levels: LogLevelDict):
        """
        :param log_levels: Dict with the registered log levels.
        :type log_levels: LogLevelDict
        """

        self.log_levels = log_levels

        self.loggers: list['Logger']             = []
        self.groups: list['Group']               = []
        self.groups_by_name: dict[str, 'Group']  = {}

    def enable_all_loggers(self):
        """Enables all loggers pinned to the logging context."""

        for l in self.loggers:
            l.enable()

    def disable_all_loggers(self):
        """Disables all loggers pinned to the logging context."""

        for l in self.loggers:
            l.disable()

    def enable_all_groups(self):
        """Enables all groups pinned to the logging context."""

        for g in self.groups:
            g.enable()

    def disable_all_groups(self):
        """Disables all groups pinned to the logging context."""

        for g in self.groups:
            g.disable()

    def get_level_offset(self):
        return len(max(self.log_levels, key=len))

    def get_logger_name_offset(self):
        return 0 if len(self.loggers) == 0 else len(max(self.loggers, key=lambda g: len(g.name)).name)

    def get_group_name_offset(self):
        return 0 if len(self.groups) == 0 else len(max(self.groups, key=lambda g: len(g.name_path)).name_path)

    @lru_cache(10)
    def log_level(self, level: LogLevel, context_level: str | int):

        if isinstance(level, LogOnlyLevels):
            if isinstance(context_level, int):
                context_level = self.log_levels.keys()[self.log_levels.values().index(context_level)]

            return level.log_level(context_level)

        elif isinstance(level, str):
            if isinstance(context_level, str):
                context_level = self.log_levels[context_level]

            return self.log_levels[level] <= context_level
        elif isinstance(level, int):
            if isinstance(context_level, str):
                context_level = self.log_levels[context_level]

            return level <= context_level



"""This module contains utilities of the library.

.. important::
    Due to the library's import system, if you import `pyrolog` by this code:

    .. code-block:: python

        import pyrolog

    You must use this as:

    .. code-block:: python

        pyrolog.utils.make_new_log_level(...)

    As example.
"""

import inspect

from datetime import datetime

from .logging_context import LoggingContext
from .defaults import DEFAULT_LOGGING_CONTEXT, MAXIMUM_TIME_FORMAT_STRING_FILENAME_SAFE
from .formatters import PlainFormatter, defined_formatters

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from .logger import Logger


__all__ = ['make_logger_binding', 'make_new_log_level', 'update_logger_name_offset',
           'update_group_name_offset', 'get_filename_timestamp']


def make_logger_binding(level: str) -> Callable:
    """Makes function-bind for given level.

    :param level: Level.
    :type level: str

    :returns: Function-bind for given level.
    :rtype: Callable
    """

    def f(self: 'Logger', message: str, *args, exc: Exception | None = None, **kwargs):
        self.record(message, level, *args, exc=exc, stack=inspect.stack()[1], **kwargs)

    return f


def make_new_log_level(logger_class: 'Logger',
                       name: str,
                       level: int,
                       logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT
                       ):
    """Makes new log level and function-bind for it. Also adds function-bind to :class:`Logger` class.
    **Note** that `name` will be converted to **lower case**!

    :param logger_class: Class of the logger.
    :type logger_class: Logger
    :param name: Name of the new log level.
    :type name: str
    :param level: Int level of the new log level.
    :type level: int
    :param logging_context: Logging context where it will be registered.
    :type logging_context: LoggingContext
    """

    name = name.lower()
    logging_context.log_levels[name] = level

    # make function-bind for the log level
    setattr(logger_class, name, make_logger_binding(name))

    # update level offset
    for f in defined_formatters:
        if hasattr(f, 'offsets') and f.offsets and f.logging_context is logging_context:
            f.static_variables['level_offset'] = logging_context.get_level_offset()


def update_logger_name_offset(logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT):
    """Updates logger name offset.

    :param logging_context: Current logging context.
    :type logging_context: LoggingContext
    """

    for f in defined_formatters:
        if hasattr(f, 'offsets') and f.offsets and f.logging_context is logging_context:
            f.static_variables['logger_name_offset'] = logging_context.get_logger_name_offset()


def update_group_name_offset(logging_context: LoggingContext = DEFAULT_LOGGING_CONTEXT):
    """Updates group name offset.

    :param logging_context: Current logging context.
    :type logging_context: LoggingContext
    """

    for f in defined_formatters:
        if hasattr(f, 'offsets') and f.offsets and f.logging_context is logging_context:
            f.static_variables['group_name_offset'] = logging_context.get_group_name_offset()


def get_filename_timestamp(format_string=MAXIMUM_TIME_FORMAT_STRING_FILENAME_SAFE + '.log'):
    """Makes pretty filename with time and ".log" extension.

    :param format_string: Format string.
    :type format_string: str
    """

    return PlainFormatter(time_format_string=format_string).format_time(datetime.now())
